package deeper.operators;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.rapidminer.operator.IOObject;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.UserError;
import com.rapidminer.operator.io.AbstractReader;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeBoolean;
import com.rapidminer.parameter.ParameterTypeCategory;
import com.rapidminer.parameter.ParameterTypeDirectory;
import com.rapidminer.parameter.ParameterTypeInt;
import com.rapidminer.parameter.ParameterTypeList;
import com.rapidminer.parameter.ParameterTypeString;
import com.rapidminer.parameter.ParameterTypeStringCategory;
import com.rapidminer.parameter.conditions.EqualTypeCondition;
import com.rapidminer.tools.jdbc.DatabaseHandler;
import com.rapidminer.tools.jdbc.connection.ConnectionEntry;
import com.rapidminer.tools.jdbc.connection.ConnectionProvider;

import deeper.entities.Abstract;
import deeper.entities.Interaction;
import deeper.entities.Protein;
import deeper.entities.Sentence;
import deeper.entities.enums.InteractionDirection;
import deeper.entities.enums.InteractionLabel;
import deeper.entities.enums.InteractionType;
import deeper.entities.graph.Graph;
import deeper.entities.rm.io.TextDataIO;

/**
 * Operator that is responsible for various database readings.
 * Supposed to provide an AbstractsIO object as output.
 * @author tfairuzo
 *
 */
public class DBReader extends AbstractReader<IOObject> implements ConnectionProvider {

	public static String PARAMETER_CORPUS_NAME = "corpus_name";
	public static String PARAMETER_CORPORA = "corpora";
	public static String PARAMETER_PARTIAL = "partial";
	public static String PARAMETER_PROCESSED = "processed";
	public static String PARAMETER_NUMBER_OF_ABSTRACTS = "number_of_abstracts";
	
	public static String[] PROCESSED_CATEGORIES = new String[]{"yes","no","both"};
	public static int PROCESSED_YES = 0;
	public static int PROCESSED_NO = 1;
	public static int PROCESSED_BOTH = 2;
	public static Map<Integer,String> INTERACTION_TYPES = null;
	
	//public static String[] datatypes = new String[]{"interaction","sentence","abstract"};


	public DBReader(OperatorDescription description) {
		super(description, TextDataIO.class);
		if(INTERACTION_TYPES==null){
			DatabaseHandler dbHandler;
			try {
				dbHandler = DatabaseHandler.getConnectedDatabaseHandler(this);
				Statement selectInteractionTypes = dbHandler.createStatement(false);
				ResultSet r = selectInteractionTypes.executeQuery("SELECT id, name FROM interaction_types");
				INTERACTION_TYPES = new HashMap<Integer,String>();
				while(r.next()){
					INTERACTION_TYPES.put(r.getInt("id"), r.getString("name"));
				}
			} catch (OperatorException e) {
				e.printStackTrace();
			} catch (SQLException e) {
				e.printStackTrace();
			}
		}
	}

	//method that is called when the operator is invoked
	@Override
	public IOObject read() throws OperatorException {
		IOObject result = null;
		try {
			DatabaseHandler dbHandler = DatabaseHandler.getConnectedDatabaseHandler(this);
			//read query parameters
			//compile a piece that selects appropriate corpora
			StringBuffer corpQuery = new StringBuffer("(");
			for( String[] corpus : getParameterList(PARAMETER_CORPORA)){
				corpQuery.append("abstracts.source='"+corpus[1]+"' OR ");
			}
			if(corpQuery.length()>1){
				corpQuery.setLength(corpQuery.length()-3);
				corpQuery.append(")");
			}else{
				corpQuery.setLength(0);
			}
			
			//filter on processed flag
			String processed = getParameterAsString(PARAMETER_PROCESSED);
			String parsed = " ";
			if(processed.equals(PROCESSED_CATEGORIES[PROCESSED_YES])){
				parsed = " AND sentences.processed=1 ";
			}else if(processed.equals(PROCESSED_CATEGORIES[PROCESSED_NO])){
				parsed = " AND sentences.processed=0 ";
			}			
			String cond1 = corpQuery.length()==0?"":" "+corpQuery.toString();
			//possibly limit the number of abstracts to read
			String cond2 = getParameterAsBoolean(PARAMETER_PARTIAL)?" LIMIT "+getParameterAsString(PARAMETER_NUMBER_OF_ABSTRACTS):"";
			PreparedStatement selectAbstracts = dbHandler.createPreparedStatement("SELECT id, external_id, source, text, title, date_published, journal FROM abstracts WHERE "+cond1+" GROUP BY abstracts.id "+cond2, false);
			PreparedStatement selectSentences = dbHandler.createPreparedStatement("SELECT id, abstract_id, sentence_number, sentence, mod_sentence, position, length, dependency_tree, parse_tree FROM sentences WHERE sentences.abstract_id=? "+parsed, false);
			PreparedStatement selectInteractions = dbHandler.createPreparedStatement("SELECT id, p1_id, p1_offset, p1_length, p2_id, p2_offset, p2_length, interaction_type_id, confidence, source FROM interaction_predictions WHERE abstract_id=? AND sentence_id=?", false);
			//PreparedStatement selectProteins = dbHandler.createPreparedStatement("SELECT name FROM entity_names WHERE entity_id=?", false);
			
			//read the whole data structure including sentences, proteins and interactions for every abstract
			List<Abstract> abstracts = readAbstractsWithContent(selectAbstracts, selectSentences, selectInteractions);
			result = new TextDataIO<Abstract>(abstracts);
			dbHandler.disconnect();
		} catch (SQLException e) {
			throw new UserError(this, e, 304, e.getMessage());
		}
	    return result;
	}

	//method that actually reads the data from database into java object realm
	//it tries to populate the data as good as it can, i.e. if there is a complete data that includes
	//abstract, sentences with parse trees and interactions, it will read all of it. If e.g.
	//interactions are missing, it would read just sentences for a given abstract.
	private List<Abstract> readAbstractsWithContent(PreparedStatement selectAbstracts, PreparedStatement selectSentences, PreparedStatement selectInteractions) throws SQLException{
		List<Abstract> abstracts = new ArrayList<Abstract>(); 
		ResultSet rs = selectAbstracts.executeQuery();
		List<Integer> abstSqlIds = new ArrayList<Integer>();
		while(rs.next()){
			Abstract abst = new Abstract();
			abst.setDb_id(rs.getInt("id"));
			abst.corpus = rs.getString("source");
			abst.journalTitle = rs.getString("journal");
			abst.setExt_id(rs.getString("external_id"));
			abst.setTitle(rs.getString("title"));
			abst.text = rs.getString("text");
			abst.dateWritten = rs.getDate("date_published");
			abstracts.add(abst);
			abstSqlIds.add(abst.getDb_id());
		}
		selectAbstracts.close();
		int k = 0;
		for(Abstract abst: abstracts){
			int abst_sql_id = abstSqlIds.get(k++);
			selectSentences.setInt(1,abst_sql_id);
			//selectSentences.setString(1, abst.pmId);
			//selectSentences.setString(2, abst.corpus);
			rs = selectSentences.executeQuery();
			while(rs.next()){
				//Sentence s = new Sentence(abst, rs.getInt("sentence_number"), rs.getString("sentence"));
				Sentence s = new Sentence(rs.getString("sentence"));
				//s.offset = rs.getInt("position");
				//s.length = rs.getInt("length");
				s.text_mod = rs.getString("mod_sentence");
				try {
					String depTree = rs.getString("dependency_tree");
					String parseTree = rs.getString("parse_tree");
					if(depTree!=null){
						s.setDepTree(Graph.deserialize(new BufferedReader(new StringReader(depTree))));
					}
					if(parseTree!=null){
						s.setParseTree(Graph.deserialize(new BufferedReader(new StringReader(parseTree))));
					}
				} catch (IOException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}

				//read proteins and interactions
				s.setDb_id(rs.getInt("id"));
				s.abst_db_id = abst_sql_id;
				selectInteractions.setInt(1, abst_sql_id);
				selectInteractions.setInt(2, s.getDb_id());
				ResultSet rsi = selectInteractions.executeQuery();
				List<Protein> proteins = new ArrayList<Protein>();
				while(rsi.next()){
					Interaction i = new Interaction(s);
					i.setDb_id(rsi.getInt("id"));
					i.abst_db_id = abst_sql_id;
					i.sentence_db_id = s.getDb_id();
					Protein p1 = new Protein();
					p1.setDb_id(rsi.getInt("p1_id"));
					p1.start_offset = rsi.getInt("p1_offset");
					p1.end_offset = p1.start_offset + rsi.getInt("p1_length");
					p1.name = s.text.substring(p1.start_offset,p1.end_offset);
					
					//ensure that proteins in the sentence list and proteins in interaction contains the same objects,
					//i.e. that p==p1 as opposed to merely p.equals(p1).
					Protein p = find(proteins, p1);
					if(p == null){
						proteins.add(p1);
					}else{
						p1 = p;
					}
					
					Protein p2 = new Protein();
					p2.setDb_id(rsi.getInt("p2_id"));
					p2.start_offset = rsi.getInt("p2_offset");
					p2.end_offset = p2.start_offset + rsi.getInt("p2_length");
					p2.name = s.text.substring(p2.start_offset,p2.end_offset);
					
					p = find(proteins, p2);
					if(p == null){
						proteins.add(p2);
					}else{
						p2 = p;
					}
					
					//we assume that proteins are stored in the order of appearance in the sentence,
					//while in the interactions_table they are stored in the order <agent, subject>
					if(p1.compareTo(p2)>0){
						i.protein1 = p2;
						i.protein2 = p1;
						i.direction = InteractionDirection.INVERSE;
					}else{
						i.protein1 = p1;
						i.protein2 = p2;
						i.direction = InteractionDirection.STRAIGHT;
					}
					String inter_type_str = INTERACTION_TYPES.get(rsi.getInt("interaction_type_id"));
					InteractionType type = new InteractionType(rsi.getInt("interaction_type_id"), inter_type_str);
					i.type = type;
					i.confidence = rsi.getFloat("confidence");
					i.source = rsi.getString("source");
					if(i.source.equals("annotation")||i.source.equals("xml")){
						if(i.confidence==1){
							i.label = InteractionLabel.POSITIVE;
						}else if(i.confidence==0){
							i.label = InteractionLabel.NEGATIVE;
						}
					}
					s.addProtein(p1);
					s.addProtein(p2);
					s.addInteraction(i);
				}
				rsi.close();
				if(s.text_mod!=null){
					s.rewrite();
				}				
				abst.addSentence(s);
			}
			rs.close();
		}
		return abstracts;
	}
	
	private Protein find(List<Protein> list, Protein p){
		for(Protein pr: list){
			if(pr.equals(p)){
				return pr;
			}
		}
		return null;
	}
	
	//abstract is not filled entirely
	/*private List<Sentence> mapSentences(ResultSet data) throws SQLException{
		List<Sentence> result = new ArrayList<Sentence>();
		Abstract currentAbst = null; 
		while(data.next()){
			String abstId = data.getString("abstId");
			String corpus = data.getString("source");
			String text = data.getString("text");
			int num = data.getInt("num");
			if(currentAbst==null || !currentAbst.pmId.equals(abstId) || !currentAbst.corpus.equals(corpus)){
	        	currentAbst = new Abstract();
	        	currentAbst.corpus = corpus;
	        	currentAbst.pmId = abstId;
	        }
			//currentAbst.text = currentAbst.text.concat(text);
			Sentence s = new Sentence(currentAbst, num, text);
			//s.sqlAbstractId = data.getInt("sqlAbstId");
			//s.sqlSentenceId = data.getInt("sqlSentenceId");
			result.add(s);
		}
		return result;
	}
	
	private List<Interaction> mapInteractions(ResultSet data){
		List<Interaction> result = new ArrayList<Interaction>();
		Map<String, Sentence> sentenceMap = new HashMap<String, Sentence>();
		while(data.next()){
			Sentence s = sentenceMap.get(data.getString("sId"));
			if(s==null){
				s = new Sentence(data.getString("sId"), data.getString("text"));
				sentenceMap.put(s.id, s);
			}
			Interaction ir = new Interaction(s);
	        ir.id = data.getInt("id");
	        Protein p1 = new Protein();
	        Protein p2 = new Protein();
	        p1.entityId = data.getInt("entity1_id");
	        p1.name = data.getString("entity1");
	        p2.entityId = data.getInt("entity2_id");
	        p2.name = data.getString("entity2");
	        ir.protein1 = p1;
	        ir.protein2 = p2;
	        ir.interactionType = data.getString("interaction_type");
	        ir.interactionTypeId = data.getInt("interaction_type_id");
	        ir.pmId = data.getString("pubmed_id");
	        ir.sentence = data.getString("sentence");
	        ir.sentenceId = data.getInt("sentence_id");
	        ir.confidence = data.getFloat("confidence");
		}
	}*/

	@Override
	public ConnectionEntry getConnectionEntry() {
		return DatabaseHandler.getConnectionEntry(this);
	}
	
	@Override
	public List<ParameterType> getParameterTypes() {
		//add parameters from superclass
		List<ParameterType> types = super.getParameterTypes();
		//add parameters that are responsible for database connection
		types.addAll(DatabaseHandler.getConnectionParameterTypes(this));
		//in principle, queries can be passed as parameters as well, however I don't use this opportunity.
		//types.addAll(DatabaseHandler.getQueryParameterTypes(this, true));
		//following parameters set different values for sql query parameters 
		ParameterType type = new ParameterTypeList(PARAMETER_CORPORA, "List of corpora to read.", new ParameterTypeString(PARAMETER_CORPUS_NAME,
			"The name of corpus as in database", "pubmed"), new ParameterTypeString("noop", "noop",true));
		//ParameterType type = new ParameterTypeString(PARAMETER_CORPUS, "Corpora to be read", "pubmed",true);
		types.add(type);
		type = new ParameterTypeCategory(PARAMETER_PROCESSED,"Select parsed abstracts or unparsed ones or don't care",PROCESSED_CATEGORIES,PROCESSED_BOTH);
		types.add(type);
		type = new ParameterTypeBoolean(PARAMETER_PARTIAL,"Read a specified number of abstracts",true);
		types.add(type);
		type = new ParameterTypeInt(PARAMETER_NUMBER_OF_ABSTRACTS, "Number of abstracts to read",1, Integer.MAX_VALUE,100);
		type.registerDependencyCondition(new EqualTypeCondition(this, PARAMETER_PARTIAL, new String[]{"true"}, true));
		types.add(type);
		return types;
	}

}
