package deeper.operators;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;

import com.rapidminer.operator.IOObject;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.UserError;
import com.rapidminer.operator.io.AbstractWriter;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeBoolean;
import com.rapidminer.parameter.ParameterTypeCategory;
import com.rapidminer.parameter.ParameterTypeString;
import com.rapidminer.tools.LogService;
import com.rapidminer.tools.jdbc.DatabaseHandler;
import com.rapidminer.tools.jdbc.connection.ConnectionEntry;
import com.rapidminer.tools.jdbc.connection.ConnectionProvider;

import deeper.entities.Abstract;
import deeper.entities.Interaction;
import deeper.entities.Protein;
import deeper.entities.Sentence;
import deeper.entities.rm.io.TextDataIO;

/**
 * WARNING: CURRENTLY, NO CHECK FOR DUPLICATES IS MADE BEFORE INSERTING TO DB!!
 * IT IS A USER RESPONSIBILITY TO MAINTIAN UNIQUENESS OF THE RECORDS IN DB. 
 * 
 * The operator that is responsible for db writing operations.
 * Current possible scenarios are: 
 *	1. reading xml file: "insert all" - inserts abstracts, sentences, protein names(if any) and interactions
 *	2. reading pubmed: "insert all" - inserts abstracts and sentences (because sentences have no interactions at this point)
 *	3. parsing and classifying interactions: "insert_interactions" - updates sentences with parse and processed flag and inserts interactions and protein names (if any)
 *	update_sentences is assumed to be called after parsing
 *	insert_interactions is currently called after the classification task is finished, that inserts the interactions along with the confidence
 * 
 * Separate sentence update mode may be useful if we make parsing a separate process. Attention! In the current implementation, parsing only makes sense
 * in case named entities are recognized, because the sentence being parsed is actually with substitution for protein names. On the other hand,
 * protein names in the sentence are not saved in sentences table, but only saved by means of interactions. Therefore, parsing a sentence that
 * have no associated interactions makes no sense.
 *   
 * @author tfairuzo
 *
 */
public class DBWriter_old extends AbstractWriter<IOObject> implements ConnectionProvider {

	public static String PARAMETER_UPDATE = "update";
	public static String PARAMETER_OVERWRITE = "overwrite";
	//public static String PARAMETER_SOURCE = "source";
	public static final String[] MODES_LIST = new String[] { "insert_all", "insert_interactions", "update_sentences" };
	public static final int MODE_INSERT_ALL = 0;
	public static final int MODE_INSERT_INTERACTIONS = 1;
	//public static final int MODE_UPDATE_SENTENCES = 2;

	public DBWriter_old(OperatorDescription description) {
		super(description, IOObject.class);
	}

	@Override
	public IOObject write(IOObject ioobject) throws OperatorException {

		TextDataIO<Abstract> abstracts = (TextDataIO<Abstract>) ioobject;
		String mode = getParameterAsString(PARAMETER_UPDATE);
		try {
			DatabaseHandler dbHandler = DatabaseHandler.getConnectedDatabaseHandler(this);
			//mode insert interactions assumes that sentences are present in db already
			if (mode.equals(MODES_LIST[MODE_INSERT_INTERACTIONS])) {
				//for every sentence we check whether it is indeed exists in db, if not we drop a message and skip it (should have used
				// 'insert_all' mode instead).
				PreparedStatement checkSentence = dbHandler.createPreparedStatement(
						"SELECT abstracts.id as abstId, sentences.id as sentId FROM abstracts, sentences WHERE " +
						"abstracts.external_id=? AND " +
						"abstracts.source=? AND " +
						"sentences.sentence_number =? AND " +
						"sentences.abstract_id=abstracts.id", false);
				
				PreparedStatement checkInteraction = dbHandler.createPreparedStatement(
						"SELECT p1_id, p2_id FROM interaction_predictions WHERE " +
						"abstract_id=? AND " +
						"sentence_id=?", false);
				
				PreparedStatement updSent = dbHandler.createPreparedStatement(
						"UPDATE sentences SET sentence_number=?, sentence=?, mod_sentence=?, position=?, length=?, dependency_tree =?, parse_tree=?, processed=? " +
						"WHERE sentences.id = ?",
						false);
				
				PreparedStatement insertInteraction = dbHandler.getConnection().prepareStatement(insertInteractionStatement(), PreparedStatement.RETURN_GENERATED_KEYS);
				PreparedStatement updateInteraction = dbHandler.getConnection().prepareStatement(updateInteractionStatement());
				
				boolean overwrite = getParameterAsBoolean(PARAMETER_OVERWRITE);
				
				PreparedStatement selectProteins = dbHandler.createPreparedStatement(selectProteinsQuery(), false);
				PreparedStatement insertEntityName = dbHandler.getConnection().prepareStatement(insertEntityNameStatement(false),  PreparedStatement.RETURN_GENERATED_KEYS);
				for (Abstract abst : abstracts.getData()) {
					for (Sentence s : abst.getSentences()) {
						//fist check whether sentence exists, if not then continue with the next sentence
						checkSentence.setString(1, abst.getExt_id());
						checkSentence.setString(2, abst.corpus);
						checkSentence.setInt(3, s.index);
						ResultSet rs = checkSentence.executeQuery();
						int abst_sql_id = -1, sentence_sql_id = -1;
						try{
							if(rs.next()){
								abst_sql_id = rs.getInt("abstId");
								sentence_sql_id = rs.getInt("sentId");
							}else{
								throw new UserError(this,304,"Sentence "+s.getExt_id()+" was passed to an update method, however it does not exist in DB, and thus cannot be updated. " +
										"I skip this sentence in the update cycle.");
							}
							rs.close();
						}catch(UserError e){
							LogService.getRoot().log(Level.WARNING, e.getMessage());
							rs.close();
							continue;
						}
						//ok, sentence exists, go and update it
						updSent.setInt(1,s.index);
						updSent.setString(2,s.text);
						updSent.setString(3,s.text_mod);
						updSent.setInt(4,s.offset);
						updSent.setInt(5, s.length);
						if(s.getDepTree()!=null){
							updSent.setString(6, s.getDepTree().serialize());
							updSent.setInt(8,1);
						}else{
							updSent.setNull(6, java.sql.Types.VARCHAR);
							updSent.setInt(8,0);
						}
						if(s.getParseTree()!=null){
							updSent.setString(7, s.getParseTree().serialize());
						}else{
							updSent.setNull(7, java.sql.Types.VARCHAR);
						}
						updSent.setInt(9, sentence_sql_id);
						updSent.executeUpdate();
						
						// when inserting a sentence we may want to insert new protein names that this sentence may contain.
						PreparedStatement insertEntity = dbHandler.getConnection().prepareStatement("INSERT INTO entities VALUES()",PreparedStatement.RETURN_GENERATED_KEYS);
						Map<Protein, Integer> proteinMap = updateProteins(selectProteins, insertEntity, insertEntityName, s);
						insertEntity.close();
						//finally, insert interactions that correspond to this sentence
						for(Interaction i : s.getInteractions()){
							insertInteraction(checkInteraction, overwrite?updateInteraction:null, insertInteraction, i, abst_sql_id, sentence_sql_id, proteinMap);
						}						
					}
				}
				insertInteraction.close();
				checkSentence.close();
				selectProteins.close();
				insertEntityName.close();
				
				///////////////////////////
				
				/*for (Abstract abst : abstracts) {
					for (Sentence s : abst.getSentences()) {
						updSent.setString(1, s.getDepTree().serialize());
						updSent.setString(2, s.getParseTree().serialize());
						updSent.setInt(3, 1);
						updSent.setInt(4, s.index);
						updSent.setString(5, s.getCorpus());
						updSent.setString(6, s.getExtId());
						updSent.executeUpdate();
						insertProcessingTime(dbHandler, s);
					}
				}
				updSent.close();
			*/
				
			}
			
			if (mode.equals(MODES_LIST[MODE_INSERT_ALL])) {
				PreparedStatement checkAbstract = dbHandler.createPreparedStatement("SELECT id FROM abstracts WHERE external_id=? AND source=?", false);
				//PreparedStatement checkSentStatement = dbHandler.createPreparedStatement("SELECT id FROM abstracts WHERE id=?", false);

				PreparedStatement selectProteins = dbHandler.createPreparedStatement(selectProteinsQuery(), false);
				PreparedStatement insertEntityName = dbHandler.getConnection().prepareStatement(insertEntityNameStatement(true),  PreparedStatement.RETURN_GENERATED_KEYS);

				PreparedStatement insertInteraction = dbHandler.getConnection().prepareStatement(insertInteractionStatement(),  PreparedStatement.RETURN_GENERATED_KEYS);
				
				PreparedStatement updateStatement = dbHandler.getConnection().prepareStatement("INSERT INTO abstracts (external_id, source, text, title, date_added, date_published, journal) " + 
						"VALUES (?,?,?,?, NOW(), ?, ?)",  PreparedStatement.RETURN_GENERATED_KEYS);
				PreparedStatement insertSentence = dbHandler.getConnection().prepareStatement("INSERT INTO sentences (abstract_id, sentence, mod_sentence, position, length, processed, sentence_number, dependency_tree, parse_tree) " + 
						"VALUES (?,?,?,?,?,?,?,?,?)",  PreparedStatement.RETURN_GENERATED_KEYS);
				PreparedStatement insertEntity = dbHandler.getConnection().prepareStatement("INSERT INTO entities VALUES()",PreparedStatement.RETURN_GENERATED_KEYS);
				for (Abstract abst : abstracts.getData()) {
					// first check whether abstract exists
					checkAbstract.setString(1, abst.getExt_id());
					checkAbstract.setString(2, abst.corpus);
					ResultSet rs = checkAbstract.executeQuery();
					boolean abstExists = rs.next();
					rs.close();
					int abst_sql_id = -1;
					if (!abstExists) // abstract does not exist
					{
						//then insert an abstract
						if(abst.dateWritten==null){
							abst.dateWritten = new Date(0);
						}
						updateStatement.setString(1,abst.getExt_id());
						updateStatement.setString(2,abst.corpus);
						updateStatement.setString(3,abst.text);
						updateStatement.setString(4,abst.getTitle());
						updateStatement.setDate(5,new java.sql.Date(abst.dateWritten.getTime()));
						updateStatement.setString(6, abst.journalTitle);
						updateStatement.executeUpdate();
						rs = updateStatement.getGeneratedKeys();
						if(rs.next()){
							abst_sql_id = rs.getInt(1);
						}else{
							throw new SQLException("Error inserting abstract.");
						}
						rs.close();
							
						// cascade insert of all depending data
						
						//insert sentences
						for (Sentence s : abst.getSentences()) {
					
							insertSentence.setInt(1,abst_sql_id);
							insertSentence.setString(2,s.text);
							insertSentence.setString(3,s.text_mod);
							insertSentence.setInt(4,s.offset);
							insertSentence.setInt(5,s.length);
							if(s.getDepTree()!=null&&s.getParseTree()!=null){
								insertSentence.setInt(6,1);
							}else{
								insertSentence.setInt(6,0);								
							}
							insertSentence.setInt(7,s.index);
							if(s.getDepTree()!=null){
								insertSentence.setString(8,s.getDepTree().serialize());
							}else{
								insertSentence.setNull(8,java.sql.Types.VARCHAR);
							}
							if(s.getParseTree()!=null){
								insertSentence.setString(9,s.getParseTree().serialize());
							}else{
								insertSentence.setNull(9,java.sql.Types.VARCHAR);
							}
							insertSentence.executeUpdate();
							int sentence_sql_id = -1;
							rs = insertSentence.getGeneratedKeys();
							if(rs.next()){
								sentence_sql_id = rs.getInt(1);
							}else{
								throw new SQLException("Error inserting sentence.");
							}
							rs.close();
							
							insertProcessingTime(dbHandler, s);
							
							// mapping and inserting protein names
							
							Map<Protein, Integer> proteinMap = updateProteins(selectProteins, insertEntity, insertEntityName, s);
							// insert interactions
							for (Interaction i : s.getInteractions()) {
								insertInteraction(null,null,insertInteraction, i, abst_sql_id, sentence_sql_id, proteinMap);
							}
						}
					}
				}
				checkAbstract.close();
				//checkSentStatement.close();
				insertEntityName.close();
				insertInteraction.close();
				selectProteins.close();
				updateStatement.close();
				insertSentence.close();
			}

			dbHandler.disconnect();
		} catch (SQLException e) {
			throw new UserError(this, e, 304, e.getMessage());
		}
		return ioobject;
	}
	
	private String insertInteractionStatement(){
		return
				"INSERT INTO interaction_predictions (abstract_id, sentence_id, p1_id, p1_offset, p1_length, p2_id, p2_offset, p2_length, interaction_type_id, confidence, date_time, source)"
						+ "					 VALUES  (?,		      ?, 		   ?,	  ?,		 ?,			?,     ?,         ?,         1,                   ?,          NOW(),     ?)";

	}
	
	private String updateInteractionStatement(){
		return
				"UPDATE interaction_predictions SET confidence=?, source=? WHERE abstract_id=?, sentence_id=?, p1_id=?, p2_id=?";

	}
	
	private String selectProteinsQuery(){
		return "SELECT entity_id FROM entity_names WHERE name=?";
	}
	
	//a parameter is true when inserting something coming from verifies source,
	//e.g. from an xml file
	private String insertEntityNameStatement(boolean checked){
		String verified = checked?"1":"0";
		return "INSERT INTO entity_names (entity_id, name, preferred, verified) VALUES(?,?,1,"+verified+")";
	}
	
	private void insertProcessingTime(DatabaseHandler dbHandler, Sentence s) throws SQLException{
		if(s.parsingTime>0){
			/*PreparedStatement selectIds = dbHandler.createPreparedStatement(
					"SELECT sentences.id, sentences.abstract_id FROM sentences,abstracts " +
					"WHERE sentences.sentence_number = ? AND " +
					"sentences.abstract_id = abstracts.id AND " +
					"abstracts.source = ? AND " +
					"abstracts.external_id = ?"
					, false);
					*/ 
			PreparedStatement updTime = dbHandler.createPreparedStatement("INSERT INTO processing_time(sentence_id, abstract_id, time) VALUES(?,?,?)",false);
			/*selectIds.setInt(1, s.index);
			selectIds.setString(2, s.getCorpus());
			selectIds.setString(3, s.getExt_id());
			ResultSet rs = selectIds.executeQuery();
			if(!rs.next()){
				throw new SQLException("Error finding sentence.");
			}else{
				updTime.setInt(1, rs.getInt("id"));
				updTime.setInt(2, rs.getInt("abstract_id"));
				updTime.setLong(3, s.parsingTime);
				updTime.executeUpdate();
			}
			rs.close();*/
			updTime.setInt(1, s.getId);
			updTime.setInt(2, rs.getInt("abstract_id"));
			updTime.setLong(3, s.parsingTime);
			updTime.executeUpdate();
		}
	}

	/*private int getEntityId(Protein entity) throws UserError {
		DatabaseHandler dbHandler;
		try {
			int entityId;
			dbHandler = DatabaseHandler.getConnectedDatabaseHandler(this);
			PreparedStatement stmt = dbHandler.createPreparedStatement("SELECT entity_id FROM entity_names WHERE LOWER(name)=?", false);

			stmt.setString(1, entity.name.trim().toLowerCase());
			ResultSet rs = stmt.executeQuery();
			if (rs.next())
				entityId = rs.getInt("entity_id");
			else {
				PreparedStatement stmt2 = dbHandler.createPreparedStatement("INSERT INTO entities () VALUES ()", false);
				stmt2.executeUpdate();
				ResultSet rskey = stmt2.getGeneratedKeys();

				if (rskey != null && rskey.next())
					entityId = rskey.getInt(1);
				else {
					throw new Exception("Error creating new entity");
				}

				PreparedStatement stmt3 = dbHandler.createPreparedStatement("INSERT INTO entity_names (entity_id, name, preferred, verified, entity_type) " + "VALUES (?, ?, 1, 0, 'protein')", false);
				stmt3.setInt(1, entityId);
				stmt3.setString(2, entity.name);
				stmt3.executeUpdate();
			}

			dbHandler.disconnect();
			return entityId;
		} catch (Exception e) {
			throw new UserError(this, e, 304, e.getMessage());
		}
	}*/

	@Override
	public ConnectionEntry getConnectionEntry() {
		return DatabaseHandler.getConnectionEntry(this);
	}

	private Map<Protein, Integer> updateProteins(PreparedStatement selectProteins, PreparedStatement insertEntity, PreparedStatement insertEntityName, Sentence s) throws SQLException {
		Map<Protein, Integer> proteinMap = new HashMap<Protein, Integer>();
		for (Protein p : s.getProteins()) {
			if (proteinMap.containsKey(p)) {
				continue;
			}
			selectProteins.setString(1, p.name);
			ResultSet rs = selectProteins.executeQuery();
			int protein_sql_id = -1;
			if (rs.next()) {
				protein_sql_id = rs.getInt("entity_id");
				rs.close();
			} else {
				rs.close();
				insertEntity.executeUpdate();
				rs = insertEntity.getGeneratedKeys();
				if(rs.next()){
					protein_sql_id = rs.getInt(1);
				}else{
					throw new SQLException("Error inserting entity.");
				}
				rs.close();				 
				insertEntityName.setInt(1, protein_sql_id);
				insertEntityName.setString(2, p.name);
				insertEntityName.executeUpdate();
			}
			proteinMap.put(p, protein_sql_id);
		}
		return proteinMap;
	}

	private void insertInteraction(PreparedStatement checkInteraction, PreparedStatement updateInteraction, PreparedStatement insertInteraction, Interaction i, int abst_sql_id, int sentence_sql_id, Map<Protein, Integer> proteinMap) throws SQLException{
		if(checkInteraction!=null){
			try{
			checkInteraction.setInt(1,abst_sql_id);
			checkInteraction.setInt(2,sentence_sql_id);
			ResultSet rs = checkInteraction.executeQuery();
			while(rs.next()){
				int p1Id = rs.getInt("p1_id");
				int p2Id = rs.getInt("p2_id");
				if((i.getProtein1().id==p1Id)&&(i.getProtein2().id==p2Id)){
					//if the interaction exists in the db, either tell it to user or update it with the new source and prediction
					//depending on parameter PARAMETER_OVERWRITE
					rs.close();
					if(updateInteraction!=null){
						updateInteraction.setDouble(1, i.confidence);
						updateInteraction.setString(2,i.source);
						updateInteraction.setInt(3,abst_sql_id);
						updateInteraction.setInt(4,sentence_sql_id);
						updateInteraction.setInt(5,p1Id);
						updateInteraction.setInt(6,p2Id);
						updateInteraction.executeUpdate();
						return;
					}else{
						throw new UserError(this,304,"Interaction "+i.getUniqueId()+" already exists in the db! Cannot add it again.");
					}					
				}
			}
			rs.close();
			}catch(UserError e){
				LogService.getRoot().log(Level.WARNING, e.getMessage());
				return;
			}
		}
		//if the interaction is not in the db, then add it
		insertInteraction.setInt(1, abst_sql_id);
		insertInteraction.setInt(2, sentence_sql_id);
		insertInteraction.setInt(3, proteinMap.get(i.protein1));
		insertInteraction.setInt(4, i.protein1.start_offset);
		insertInteraction.setInt(5, i.protein1.name.length());
		insertInteraction.setInt(6, proteinMap.get(i.protein2));
		insertInteraction.setInt(7, i.protein2.start_offset);
		insertInteraction.setInt(8, i.protein2.name.length());
		insertInteraction.setDouble(9, i.confidence);
		insertInteraction.setString(10, i.source);
		insertInteraction.executeUpdate();
	}

	@Override
	public List<ParameterType> getParameterTypes() {
		List<ParameterType> types = super.getParameterTypes();
		types.addAll(DatabaseHandler.getConnectionParameterTypes(this));
		types.addAll(DatabaseHandler.getQueryParameterTypes(this, true));
		ParameterType type = new ParameterTypeCategory(PARAMETER_UPDATE, "Insert abstracts with all dependencies, insert interactions or update sentences", MODES_LIST, MODE_INSERT_ALL);
		types.add(type);
		type = new ParameterTypeBoolean(PARAMETER_OVERWRITE, "Overwrite the predictions set by another classifier",false);
		types.add(type);
		return types;
	}

}
