package deeper.operators;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.table.AttributeFactory;
import com.rapidminer.example.table.DataRow;
import com.rapidminer.example.table.DataRowFactory;
import com.rapidminer.example.table.MemoryExampleTable;
import com.rapidminer.example.table.NominalMapping;
import com.rapidminer.operator.OperatorChain;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.UserError;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.operator.ports.metadata.ExampleSetMetaData;
import com.rapidminer.operator.ports.metadata.GenerateNewMDRule;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeCategory;
import com.rapidminer.parameter.ParameterTypeFile;
import com.rapidminer.parameter.UndefinedParameterError;
import com.rapidminer.repository.RepositoryLocation;

import deeper.entities.Abstract;
import deeper.entities.Interaction;
import deeper.entities.Sentence;
import deeper.entities.rm.io.InteractionIO;
import deeper.entities.rm.io.TextDataIO;

/**
 * Converts a collection of interactions into an ExampleSet.
 * This operator is an OperatorChain, which means that it should contain inner operator to work
 * properly. This inner operator is responsible for putting the features that we need in ExampleSet
 * into a valueMap of an InteractionIO object. The example implementation is TreeVectorCompilerOperator. 
 * @author tfairuzo
 *
 */
public class Text2ExampleSetOperator extends OperatorChain{
	
    public static final String PARAMETER_DATAMANAGEMENT = "datamanagement";
    public static final String PARAMETER_DEPTREE_MAPPING = "deptree_mapping";
    public static final String PARAMETER_PARSETREE_MAPPING = "parsetree_mapping";
    public static final String PARAMETER_MODE = "mode";
    public static final String[] MODES = new String[]{"train","test"};
    public static final int TRAIN_MODE = 0;
    public static final int TEST_MODE = 1;
    
	private InputPort textsInput = getInputPorts().createPort("texts", TextDataIO.class);
	private OutputPort exampleSetOutput = getOutputPorts().createPort("example set");
	private OutputPort textOutput = getOutputPorts().createPort("through");
	private OutputPort innerInteractionSource = getSubprocess(0).getInnerSources().createPort("interaction");
	private InputPort innerInteractionSink = getSubprocess(0).getInnerSinks().createPort("interaction", InteractionIO.class);

	public Text2ExampleSetOperator(OperatorDescription description) {
		super(description,"Feature Converter");
		
		getTransformer().addRule(new GenerateNewMDRule(exampleSetOutput, new ExampleSetMetaData()));
	}
	
	public void doWork() throws OperatorException{
		DataRowFactory factory = new DataRowFactory(getParameterAsInt(PARAMETER_DATAMANAGEMENT), '.');
        MemoryExampleTable resultSet = null;
        String mode = getParameterAsString(PARAMETER_MODE);
        TextDataIO<Abstract> abstsIO = textsInput.getData();
		//iterate through all interaction instances
		for(Abstract abst: abstsIO.getData()){
			for(Sentence s: abst.getSentences()){
				for(Interaction i: s.getInteractions()){
					InteractionIO iIO = new InteractionIO(i); 
					innerInteractionSource.deliver(iIO);
					//and pass every instance to the inner process that is responsible
					//for setting features for this instance
		        	getSubprocess(0).execute();
		        	InteractionIO processedI = innerInteractionSink.getDataOrNull();
		        	if(processedI==null){
		        		continue;
		        	}
		        	if (resultSet == null) {
                    	//initialize the structure of underlying data table
                        resultSet = initExampleTable(processedI, mode);
                    }
                    //add example that conforms to the structure of data table
                    addExample(resultSet, processedI, factory);
				}
			}
		}
		
		if(resultSet == null){
			throw new UserError(this, 117);
		}
		
		if(mode.equals(MODES[TRAIN_MODE])){//serialize nominal mappings of training sets
			File depMapping = getProcess().resolveFileName(getParameterAsString(PARAMETER_DEPTREE_MAPPING)).getAbsoluteFile();
        	File parseMapping = getProcess().resolveFileName(getParameterAsString(PARAMETER_PARSETREE_MAPPING)).getAbsoluteFile();
        	Attribute att = resultSet.findAttribute("depTree");
        	try{
				serializeMapping(att.getMapping(), depMapping);
			}catch(IOException e){
				throw new OperatorException("Cannot serialize depTree attribute nominal mapping", e);
			}
			att = resultSet.findAttribute("parseTree");
        	try{
				serializeMapping(att.getMapping(), parseMapping);
			}catch(IOException e){
				throw new OperatorException("Cannot serialize parseTree attribute nominal mapping", e);
			}
        }
		
		//make an example set from the data table with a special attribute "label"
		ExampleSet exampleSet = null;
		Map<Attribute,String> special = new HashMap<Attribute, String>();
		//male "id" a special attribute (it would not be used for classification)
		
		//try{
			for(Attribute att: resultSet.getAttributes()){
				if(att.getName().equals("id")){
					special.put(att, "id");
				}else if(att.getName().equals("label")){
					special.put(att, "label");
				}else if(att.getName().equals("prediction")){
					special.put(att, "prediction");
				}
			}
			//try to find a label and set it a special attribute
			/*Attribute label = resultSet.findAttribute("label");
			special.put(label, "label");
			exampleSet = resultSet.createExampleSet(special);
			
			Attribute prediction = resultSet.findAttribute("prediction");
			special.put(label, "label");
			special.put(prediction, "prediction");
			*/
			
			exampleSet = resultSet.createExampleSet(special);
			
			
		/*}catch(UserError e){
			if(e.getCode()==111){
				//if failed because attribute label not found, we are possibly in production mode,
				//so proceed without label attribute.
				exampleSet = resultSet.createExampleSet(special);
			}else{
				throw new OperatorException("",e);
			}
		}*/
		//another piece of magic that seems to be required in order to make ids work
		exampleSet.remapIds();
        exampleSetOutput.deliver(exampleSet);
        textOutput.deliver(abstsIO);
	}
	
	private void serializeMapping(NominalMapping mapping, File file) throws IOException{
		ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(file));
		out.writeObject(mapping);
		out.close();
	}
	
	private NominalMapping readMapping(File file) throws OperatorException{
		NominalMapping mapping = null; 
		try {
			ObjectInputStream ois = new ObjectInputStream(new FileInputStream(file));
			mapping = (NominalMapping) ois.readObject();
			ois.close();
		} catch (FileNotFoundException e) {
			throw new OperatorException("Cannot find the nominal mapping file",e);
		} catch (IOException e) {
			throw new OperatorException("Error reading the nominal mapping",e);
		} catch (ClassNotFoundException e) {
			throw new OperatorException("Deserialization problems",e);
		}
		return mapping;
	}
	
	//the method defines the structure of the ExampleTable, i.e. which features (columns) are going to be there
	private MemoryExampleTable initExampleTable(InteractionIO inter, String mode) throws OperatorException{
        // creating table and attributes
        // if new attributes of interactions should be considered, add it here
        MemoryExampleTable table = new MemoryExampleTable();
        List<Attribute> metaDataAttributes = new LinkedList<Attribute>();
        Iterator<String> attIterator = inter.getNameIterator();
        File depMapping = null, parseMapping = null;
        try {
			depMapping = getProcess().resolveFileName(getParameterAsString(PARAMETER_DEPTREE_MAPPING)).getAbsoluteFile();
			parseMapping = getProcess().resolveFileName(getParameterAsString(PARAMETER_PARSETREE_MAPPING)).getAbsoluteFile();
		} catch (UndefinedParameterError e) {
			e.printStackTrace();
		} 
        
        while (attIterator.hasNext()) {
            String name = attIterator.next();
            Attribute attribute = AttributeFactory.createAttribute(name, inter.getValueType(name));
            //a bit of magic: the thing is that every nominal (string) attribute in RM is mapped internally to
            //a double value. The problem is that when classifier builds a model, it uses one mapping,
            //while when the mode is applied in another dataset, it may have another mapping. To ensure
            //that the mapping is correct, we need to store mapping of training dataset, and restore it for 
            //every test dataset.
            //Update: RM seem to do it automatically, so maybe we don't need it.
            if(mode.equals(MODES[TEST_MODE])&&name.equals("depTree")&&depMapping!=null){
            	attribute.setMapping(readMapping(depMapping));
            }
            if(mode.equals(MODES[TEST_MODE])&&name.equals("parseTree")&&parseMapping!=null){
            	attribute.setMapping(readMapping(depMapping));
            }
            metaDataAttributes.add(attribute);
        }
        //metaDataAttributes.add(AttributeFactory.createAttribute("id", Ontology.NOMINAL));
        table.addAttributes(metaDataAttributes);
        return table;
    }

	//adding an example to a dataset with established structure
    private void addExample(MemoryExampleTable resultSet, InteractionIO inter, DataRowFactory factory) {
        DataRow row = factory.create(resultSet.getAttributeCount());
        for (Attribute att : resultSet.getAttributes()) {
            if (att.isNominal()) {
            	//System.out.println(inter.getInteraction().getUniqueId());
                String value = inter.getValueAsString(att.getName());
                row.set(att, att.getMapping().mapString(value));
            }
            if (att.isNumerical()) {
          		row.set(att, inter.getValueAsDouble(att.getName()));
            }
        }
        resultSet.addDataRow(row);
    }

	
    @Override
    public List<ParameterType> getParameterTypes() {
        List<ParameterType> types = super.getParameterTypes();

        types.add(new ParameterTypeCategory(PARAMETER_DATAMANAGEMENT, "Determines, how the data is represented internally.", DataRowFactory.TYPE_NAMES, DataRowFactory.TYPE_DOUBLE_SPARSE_ARRAY));
        types.add(new ParameterTypeFile(PARAMETER_DEPTREE_MAPPING, "The file that contains the nominal mapping of depencency trees used in training", "nmp",true));
        types.add(new ParameterTypeFile(PARAMETER_PARSETREE_MAPPING, "The file that contains the nominal mapping of parse trees used in training", "nmp",true));
        types.add(new ParameterTypeCategory(PARAMETER_MODE, "Which type of example set to generate", MODES, TEST_MODE));
        return types;
    }


}
