package deeper.operators.converters;

//the getSigmoidParameters code is taken from 
//A Note on Platts Probabilistic Outputs for Support Vector Machines. Hsuan-Tien Lin, Chih-Jen Lin, Ruby C. Weng
public class PlattProbabilityEstimator {
	
	public double a,b;
	
	public PlattProbabilityEstimator(double[] predictions, boolean[] labels, int numPos, int numNeg){
		double[] params = getSigmoidParameters(predictions, labels, numPos, numNeg);
		a = params[0];
		b = params[1];
	}
	
	public PlattProbabilityEstimator(double a, double b){
		this.a = a;
		this.b = b;
	}
	
	public double getProbability(double prediction){
		double fApB = prediction*a+b;
		if (fApB >= 0)
			return Math.exp(-fApB)/(1.0+Math.exp(-fApB));
		else
			return 1.0/(1+Math.exp(fApB)) ;
	}
	
	//returns sigmoid parameters a and b
	private double[] getSigmoidParameters(double[] predictions, boolean[] labels, int numPos, int numNeg){
		//deci = array of SVM decision values
		//label = array of booleans: is the example labeled +1?
		//prior1 = number of positive examples
		//prior0 = number of negative examples

		//Parameter setting
		int maxiter=100; //Maximum number of iterations
		double minstep=1E-10D; //Minimum step taken in line search
		double sigma=1E-12D; //Set to any value > 0
		//Construct initial values: target support in array t,
		// initial function value in fval
		double hiTarget=(numPos+1.0)/(numPos+2.0); 
		double loTarget=1/(numNeg+2.0);
		int len=numPos+numNeg; // Total number of data
		double[] t = new double[len];
		for(int i = 0; i<len;i++){
			if (labels[i]){
				t[i]=hiTarget;
			}else{
				t[i]=loTarget;
			}
		}
		double A=0.0, B=Math.log((numNeg+1.0)/(numPos+1.0)), fval=0.0;
		for(int i = 0; i<len; i++) {
			double fApB=predictions[i]*A+B;
			if (fApB >= 0){
				fval += t[i]*fApB+Math.log(1+Math.exp(-fApB));
			}else{
				fval += (t[i]-1)*fApB+Math.log(1+Math.exp(fApB));
			}
		}
		for(int it = 0; it<maxiter;it++) {
			//Update Gradient and Hessian (use H = H + sigma I)
			double h11,h22, h21,g1,g2;
			h11 = h22 = sigma;
			h21 = g1 = g2 = 0.0;
			for(int i = 0;i<len;i++) {
				double fApB=predictions[i]*A+B;
				double p,q;
				if (fApB >= 0){
					p=Math.exp(-fApB)/(1.0+Math.exp(-fApB));
					q=1.0/(1.0+Math.exp(-fApB));
				}else{
					p=1.0/(1.0+Math.exp(fApB));
					q=Math.exp(fApB)/(1.0+Math.exp(fApB));
				}		
				double d2=p*q;
				h11 += predictions[i]*predictions[i]*d2;
				h22 += d2;
				h21 += predictions[i]*d2;
				double d1=t[i]-p;
				g1 += predictions[i]*d1;
				g2 += d1;
			}
			if (Math.abs(g1)<1E-5D && Math.abs(g2)<1E-5D) //Stopping criteria
				break;
			//Compute modified Newton directions
			double det=h11*h22-h21*h21;
			double dA=-(h22*g1-h21*g2)/det, dB=-(-h21*g1+h11*g2)/det;
			double gd=g1*dA+g2*dB;
			int stepsize=1;
			while (stepsize >= minstep){ //Line search
				double newA=A+stepsize*dA, newB=B+stepsize*dB, newf=0.0;
				for(int i = 0; i<len; i++ ){
					double fApB=predictions[i]*newA+newB;
					if (fApB >= 0){
						newf += t[i]*fApB+Math.log(1+Math.exp(-fApB));
					}else{
						newf += (t[i]-1)*fApB+Math.log(1+Math.exp(fApB));
					}
				}
				if (newf<fval+0.0001*stepsize*gd){
					A=newA;
					B=newB;
					fval=newf;
					break; //Sufficient decrease satisfied
				}else{
					stepsize /= 2.0;
				}
			}
			if (stepsize < minstep){
				System.out.println("Line search fails");
				break;
			}
			if (it >= maxiter)
				System.out.println("Reaching maximum iterations");
		}		
		return new double[]{A,B};
	}
}
