package deeper.operators.converters;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Attributes;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.table.AttributeFactory;
import com.rapidminer.example.table.ExampleTable;
import com.rapidminer.example.table.NominalMapping;
import com.rapidminer.tools.Ontology;

import deeper.entities.Abstract;
import deeper.entities.Interaction;
import deeper.entities.Sentence;
import deeper.entities.TextData;
import deeper.entities.enums.InteractionLabel;
import deeper.entities.graph.Graph;

//a collection of converters for SVMLight format
//warning: as this format does not include example ids, order matters
//i.e. Abstracts in abstracts2Moschitti and in predictions2Abstracts should be accessed in the same order.
public class SVMLightConverters {

	// converts a collection of abstracts to an SVMLight format, that
	// includes lexical, POS and syntactic paths
	// returns labels of the set as an array
	public static InteractionLabel[] abstracts2Moschitti(TextData<Abstract> absts, String file, int vectorType) throws IOException {
		List<InteractionLabel> result = new ArrayList<InteractionLabel>();
		FileWriter writer = new FileWriter(file);
		for (Abstract abst : absts) {
			for (Sentence s : abst.getSentences()) {
				for (Interaction i : s.getInteractions()) {
					if (i.getDepPath() != null && i.getDepPath().getVertices().size() != 0) {
						InteractionLabel lab = i.label;
						if (i.label != null) {
							result.add(i.label);
						} else {
							result.add(InteractionLabel.UNDECIDED);
							lab = InteractionLabel.UNDECIDED;
						}
						writeExample(lab, vectorType, i.getDepPath(), writer);
					}
				}
			}
		}
		writer.close();
		return result.toArray(new InteractionLabel[0]);
	}

	// converts a collection of abstracts to an SVMLight format, that
	// includes lexical, POS and syntactic paths
	// returns labels of the set as an array
	public static InteractionLabel[] abstracts2Moschitti(ExampleSet exSet, String file, int vectorType) throws IOException {
		List<InteractionLabel> result = new ArrayList<InteractionLabel>();
		FileWriter writer = new FileWriter(file);
		Iterator<Example> it = exSet.iterator();
		while (it.hasNext()) {
			Example ex = it.next();
			Graph depPath = null;
			InteractionLabel lab = null;
			for (Iterator<Attribute> i = ex.getAttributes().allAttributes();i.hasNext();) {
				Attribute att = i.next();
				if (att.getName().equals("depTree")) {
					depPath = Graph.deserialize(new BufferedReader(new StringReader(ex.getNominalValue(att))));
				}
				if (att.getName().equals("label")) {
					lab = InteractionLabel.valueOf(ex.getNominalValue(att));
				}
			}
			if (depPath != null && depPath.getVertices().size() != 0) {
				if (lab != null) {
					result.add(lab);
				} else {
					result.add(InteractionLabel.UNDECIDED);
					lab = InteractionLabel.UNDECIDED;
				}
				writeExample(lab, vectorType, depPath, writer);
			}
		}
		writer.close();
		return result.toArray(new InteractionLabel[0]);
	}

	private static void writeExample(InteractionLabel label, int vectorType, Graph path, FileWriter writer) throws IOException {
		if (label == InteractionLabel.POSITIVE) {
			writer.write("+1 ");
		} else if (label == InteractionLabel.NEGATIVE) {
			writer.write("-1 ");
		} else if (label == InteractionLabel.UNDECIDED) {
			writer.write("0 ");
		}
		switch (vectorType) {
		case 0:
			writer.write("|BT| " + MochittiFormat.convertLex(path) + " |ET|");
			break;
		case 1:
			writer.write("|BT| " + MochittiFormat.convertPos(path) + " |ET|");
			break;
		case 2:
			writer.write("|BT| " + MochittiFormat.convertSyn(path) + " |ET|");
			break;
		case 3:
			writer.write("|BT| " + MochittiFormat.convertLex(path) + " |BT|" + MochittiFormat.convertPos(path) + " |ET|");
			break;
		case 4:
			writer.write("|BT| " + MochittiFormat.convertLex(path) + " |BT|" + MochittiFormat.convertSyn(path) + " |ET|");
			break;
		case 5:
			writer.write("|BT| " + MochittiFormat.convertLex(path) + " |BT|" + MochittiFormat.convertPos(path) + " |BT| " + MochittiFormat.convertSyn(path) + " |ET|");
			break;
		case 6:
			writer.write("|BT| " + MochittiFormat.convertPos(path) + " |BT| " + MochittiFormat.convertSyn(path) + " |BT|");
			break;
		default:
			writer.write("|BT| " + MochittiFormat.convertLex(path) + " |BT|" + MochittiFormat.convertPos(path) + " |BT| " + MochittiFormat.convertSyn(path) + " |ET|");
			break;

		}
		writer.write("\n");
	}

	public static double[] predictions2ExampleSet(ExampleSet exSet, String file) throws IOException {
		// List<Double> result = new ArrayList<Double>();
		// BufferedReader reader = new BufferedReader(new FileReader(file));
		
		createPredictedLabel(exSet, exSet.getAttributes().getLabel());
		
		double[] result = readPredictions(file);
		int k = 0;
		Iterator<Example> it = exSet.iterator();
		NominalMapping mapping = exSet.getAttributes().get("depTree").getMapping();
		Attribute att = exSet.getAttributes().get("depTree");
		while (it.hasNext()) {
			Example ex = it.next();
			Graph depPath = Graph.deserialize(new BufferedReader(new StringReader(mapping.mapIndex((int)ex.getValue(att)))));
			if (depPath != null && depPath.getVertices().size() != 0) {
				setPrediction(ex,result[k++]);
			}
		}
		// double[] final_result = new double[result.size()];
		// for (int i = 0; i < result.size(); i++) {
		// final_result[i] = result.get(i);
		// }
		return result;
	}

	// adds the classifier output to corresponding interactions
	// returns an array of predictions
	public static double[] predictions2Abstracts(TextData<Abstract> absts, String file) throws IOException {
		// List<Double> result = new ArrayList<Double>();
		// BufferedReader reader = new BufferedReader(new FileReader(file));
		double[] result = readPredictions(file);
		int k = 0;
		for (Abstract abst : absts) {
			for (Sentence s : abst.getSentences()) {
				for (Interaction i : s.getInteractions()) {
					if (i.getDepPath() != null && i.getDepPath().getVertices().size() != 0) {
						i.confidence = result[k++];
						if (i.confidence > 0) {
							i.prediction = InteractionLabel.POSITIVE;
						} else {
							i.prediction = InteractionLabel.NEGATIVE;
						}
					}
				}
			}
		}
		// double[] final_result = new double[result.size()];
		// for (int i = 0; i < result.size(); i++) {
		// final_result[i] = result.get(i);
		// }
		return result;
	}

	public static double[] readPredictions(String file) throws IOException {
		List<Double> result = new ArrayList<Double>();
		BufferedReader reader = new BufferedReader(new FileReader(file));
		String s = "";
		while ((s = reader.readLine()) != null) {
			result.add(Double.parseDouble(s));
		}
		double[] final_result = new double[result.size()];
		for (int i = 0; i < result.size(); i++) {
			final_result[i] = result.get(i);
		}
		return final_result;
	}
	
	//methods stealed from RM to mimic Model behaviour when necessary 

	private static void setPrediction(Example example, double prediction) {
		Attribute predLabel = example.getAttributes().getPredictedLabel();
		int index = prediction > 0 ? predLabel.getMapping().getPositiveIndex() : predLabel.getMapping().getNegativeIndex();
		example.setValue(predLabel, index);
		example.setValue(example.getAttributes().get("confidence"), prediction);
	}
	
	public static Attribute createPredictedLabel(ExampleSet exampleSet, Attribute label) {
		// create and add prediction attribute
		Attribute predictedLabel = AttributeFactory.createAttribute(label, Attributes.PREDICTION_NAME);
		predictedLabel.clearTransformations();
		ExampleTable table = exampleSet.getExampleTable();
		table.addAttribute(predictedLabel);
		exampleSet.getAttributes().setPredictedLabel(predictedLabel);
        
		// create and add confidence attributes for nominal labels
		Attribute confidence = AttributeFactory.createAttribute("confidence", Ontology.REAL);
		table.addAttribute(confidence);
		exampleSet.getAttributes().setSpecialAttribute(confidence, "other");
		return predictedLabel;
	}
}
