package deeper.parser;

import deeper.entities.graph.Graph;
import deeper.entities.graph.Node;
import edu.stanford.nlp.ling.Sentence;
import edu.stanford.nlp.ling.TaggedWord;
import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.trees.TreeGraphNode;
import edu.stanford.nlp.trees.TypedDependency;
import edu.stanford.nlp.ling.CoreAnnotations.CopyAnnotation;
import edu.stanford.nlp.ling.CoreAnnotations.IndexAnnotation;
import java.util.Collection;

/**
 * Factory converts Stanford Parser parse and dependency tree representations
 * to the JUNG graph format.
 * Note, that the only JUNG dependency is return types
 * All other specific code is isolated in separate builders
 * @author tfayruzo
 *
 */
public class StanfordGraphFactory implements GraphFactory<Collection<TypedDependency>, Tree, Sentence<TaggedWord>> {

    /**
     *
     * @param depTree - Stanford Parser dependency tree
     * @return directed graph (with tree skeleton) in JUNG format
     */
    @Override
    public Graph getDepGraph(Collection<TypedDependency> depTree) {
        DependencyTreeBuilder builder = new DependencyTreeBuilder();
        builder.init();
        for (TypedDependency dependency : depTree) {
            Node gov = makeNode(dependency.gov());
            Node dep = makeNode(dependency.dep());
            String edgeLabel = dependency.reln().getShortName();
            builder.add(gov,dep,edgeLabel);
        }
        Graph depGraph = builder.getDepTree();
        //depGraph.initRoot();
        return depGraph;
    }

    /**
     *
     * @param parseTree - Stanford Parser dependency tree
     * @return directed tree graph in JUNG format
     */
    @Override
    public Graph getParseGraph(Tree parseTree) {
        //internalId = -1;
        ParseTreeBuilder builder = new ParseTreeBuilder();
        builder.init(parseTree.label().value());
        buildTree(parseTree, builder);
        return builder.getParseTree();
    }

    @Override
    public StructuredSentence getSentence(Sentence<TaggedWord> sentence) {
        return new StructuredSentence(sentence);
    }

    //---------non-interface methods-------
    public Graph addPOS(Graph g, StructuredSentence sentence) {
        for (Node v : g.getVertices()) {
            String pos = sentence.getPos(v.index);
            v.pos = pos;
        }
        return g;
    }

    //---------private methods-------------
    private void buildTree(edu.stanford.nlp.trees.Tree tree, ParseTreeBuilder builder) {
       Node current = builder.getCurrent();
        for (edu.stanford.nlp.trees.Tree child : tree.children()) {
            Node jChild = null;
            if (child.isLeaf()) {
                jChild = builder.makeLeaf(child.label().value());
            } else {
                jChild = builder.makeInnerNode(child.label().value());
            }
            builder.addEdge(current, jChild);
            builder.moveParent(jChild);
            buildTree(child, builder);
            builder.moveParent(current);
        }
    }

    private Node makeNode(TreeGraphNode aNode){
        Node n = new Node();
        //in the Stanford parser words are indexed from 1 which is inconvenient for us
        Integer index = aNode.label().get(IndexAnnotation.class);
        n.word = aNode.label().toString("value");
        if (aNode.label().get(CopyAnnotation.class)!=null) {
            index += 1000;
        }
        n.index = index-1;
        return n;
    }
}
