package deeper.parser;

import deeper.entities.graph.Graph;
import java.util.Collection;

import edu.stanford.nlp.ling.Sentence;
import edu.stanford.nlp.ling.TaggedWord;
import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.trees.TypedDependency;

/**
 * Implementation of Parser interface for the Stanford Parser
 * TODO: wrap log messages
 * @author tfayruzo
 *
 */
public class StanfordParser implements Parser {

    //parser wrapper is isolated in separate class just for the sake of
    //keeping parser code separately. Probably put it right here?
    private StanfordParserWrapper parser = null;
    //GraphFactory isolated to separate class to keep JUNG code
    //as much independent as possible
    private StanfordGraphFactory factory = null;

    /**
     * Acceptable arguments
     * -maxLength int
     * -dependencyType dependencies|collapsedDependencies|typedDependenciesCCprocessed
     */
    @Override
    public void init(String... args) {
        int maxLength = 0;
        for (int i = 0, size = args.length; i < size; i++) {
            if (args[i].equals("-maxLength")) {
                maxLength = Integer.parseInt(args[++i]);
            }
        }
        parser = new StanfordParserWrapper(maxLength);
        factory = new StanfordGraphFactory();
    }

    @Override
    public void parse(String sentence) throws IllegalStateException, Exception {
        if (parser == null) {
            throw new IllegalStateException("Parser should be initialized before calling parse() method.");
        }
        try {
            parser.getParse(sentence);
        } catch (UnsupportedOperationException e) {
            throw new Exception(sentence + " Nested exception message: " + e.getMessage());
        }
    }

    @Override
    public Graph getDependencyTree() throws IllegalStateException {
        Collection<TypedDependency> deps = parser.getTypedDependencies();
        if (deps == null) {
            throw new IllegalStateException("No parse() method was executed. Dependency tree not yet exists.");
        }
        //getting dependency graph
        Graph g = factory.getDepGraph(deps);
        //adding POS information
        g = factory.addPOS(g, getStructuredSentence());
        return g;
    }

    @Override
    public Graph getParseTree() throws IllegalStateException {
        Tree tree = parser.getParseTree();
        if (tree == null) {
            throw new IllegalStateException("No parse() method was executed. Parse tree not yet exists.");
        }
        return factory.getParseGraph(tree);
    }

    @Override
    public StructuredSentence getStructuredSentence() throws IllegalStateException {
        Sentence<TaggedWord> sent = parser.getSentence();
        if (sent == null) {
            throw new IllegalStateException("No parse() method was executed. Structured sentence not yet exists.");
        }
        return factory.getSentence(sent);
    }

    public static void main(String[] args) throws Exception {
        Parser parser = new StanfordParser();
        parser.init("-maxLength", "70");
        String sent = "Colorless green ideas sleep furiously.";
        parser.parse(sent);
        Graph depTree = parser.getDependencyTree();
        Graph parseTree = parser.getParseTree();
        StructuredSentence ssent = parser.getStructuredSentence();
        System.out.println(ssent.toString());
        System.out.println(depTree.serialize());
        System.out.println(parseTree.serialize());
    }
}
