package deeper.parser;

import java.util.Collection;
import java.util.List;

import edu.stanford.nlp.ling.Sentence;
import edu.stanford.nlp.ling.TaggedWord;
import edu.stanford.nlp.ling.Word;
import edu.stanford.nlp.parser.lexparser.LexicalizedParser;
import edu.stanford.nlp.process.DocumentPreprocessor;
import edu.stanford.nlp.trees.GrammaticalStructure;
import edu.stanford.nlp.trees.GrammaticalStructureFactory;
import edu.stanford.nlp.trees.Tree;
import edu.stanford.nlp.trees.TreePrint;
import edu.stanford.nlp.trees.TreebankLanguagePack;
import edu.stanford.nlp.trees.TypedDependency;

/**
 * Wrapper for the Stanford parser, encapsulates the original library and provides the 
 * basic functionality that is necessary for DEEPER.
 * 
 *  However, I will use Sentence class through the application, since acts independently
 *  of other parser parts, and provides very nice functionality for word iteration and counting. 
 * @author tfayruzo
 *
 */
public class StanfordParserWrapper {

    private LexicalizedParser lp = new LexicalizedParser("lib/englishPCFG.ser.gz");
    private String text;
    private Tree parseTree;
    private Sentence<TaggedWord> sentence;
    private GrammaticalStructure depTree;
    private Collection<TypedDependency> deps;

    public StanfordParserWrapper(int maxLength) {
        lp.setOptionFlags(new String[]{"-maxLength", Integer.toString(maxLength)});
        lp.setOptionFlags(new String[]{"-retainTmpSubcategories"});
        lp.setOptionFlags(new String[]{"-outputFormat", "penn,typedDependencies"});
        lp.setOptionFlags(new String[]{"-outputFormatOptions", "stem"});
    }

    public Collection<TypedDependency> getTypedDependencies() {
        if (!GrammaticalStructure.isConnected(deps)) {
            //_log.error("Dependency tree has more than one root");
        }
        return deps;
    }

    private GrammaticalStructure buildDependencyTree() {
        TreebankLanguagePack tlp = lp.getOp().tlpParams.treebankLanguagePack();
        GrammaticalStructureFactory factory = tlp.grammaticalStructureFactory();
        return factory.newGrammaticalStructure(parseTree);
    }

    public Tree getParse(String sentence) throws UnsupportedOperationException {
        //this.sentence = sentence;
        if (sentence == null) {
            throw new UnsupportedOperationException("Sentence is null.");
        }
        if(sentence.equals(text)){
            return parseTree;
        }
        DocumentPreprocessor dp = new DocumentPreprocessor(lp.getOp().tlpParams.treebankLanguagePack().getTokenizerFactory());
        List<Word> lst = dp.getWordsFromString((String) sentence);
        //int i = 0;
        //TODO: handle partially parsed input
        /*Pattern p = Pattern.compile("\\\\/[A-Z]+");
        for (Word word : lst) {
            Matcher m = p.matcher(word.word());
            if (m.find()) {
                String POStag = m.group().substring(3);
                lst.set(i, new TaggedWord(word.word().replace(POStag, ""), POStag));
            }
            i++;
        }
        */
        if (lp.parse(lst)) {
            parseTree = lp.getBestParse();
        }
        this.sentence = parseTree.taggedYield();
        depTree = buildDependencyTree();
        deps = depTree.typedDependenciesCCprocessed(false);
        return parseTree;
    }

    public Tree getParseTree() {
        return parseTree;
    }

    public static void main(String[] args) {
        //String sentence = "Bills on ports and immigration were submitted by Senator Brownback , Republican of Kansas";
        //String sentence = "However, the mechanism by which activin exerts its inhibitory effects are not yet understood.";
        //String sentence = "In this clinical trial, children were randomly assigned to one of three groups and received either one single oral high dose of VA (200 000 IU) on the day of admission and subsequently a placebo daily until discharge or daily oral low doses of VA (5000 IU) from admission until discharge or a placebo daily from admission until discharge.";
        String sentence = "Colorless green ideas sleep furiously.";
        StanfordParserWrapper parser = new StanfordParserWrapper(70);
        parser.getParse(sentence);
        TreePrint tp = parser.lp.getTreePrint();
        tp.printTree(parser.parseTree);
    }

    public Sentence<TaggedWord> getSentence() {
        return sentence;
    }
}